@@ -75,11 +75,13 @@ struct steve_waiter {
 	}
 };
 
+typedef std::unique_ptr<struct event, std::function<void(struct event*)>> event_ptr;
+
 struct steve_process {
 	int pid_fd{-1};
 	ssize_t tokens_held{0};
 	bool token_reserved{false};
-	std::unique_ptr<struct event, std::function<void(struct event*)>> event;
+	event_ptr pidfd_event;
 
 	~steve_process() {
 		if (pid_fd != -1)
@@ -88,9 +90,12 @@ struct steve_process {
 };
 
 struct steve_state {
+	int retval{0};
+	const char *dev_name{"steve"};
 	bool verbose;
 	uint64_t jobs;
-	uint64_t min_jobs;
+	uint64_t min_jobs{1};
+	int64_t per_process_limit;
 	double max_load_avg{-1};  /* < 0 implies no load average */
 	double load_avg;
 	int64_t tokens;
@@ -100,7 +105,7 @@ struct steve_state {
 	int loadavg_fd{-2};
 	struct timeval recheck_timeout{0, 500000};
 	bool recheck_triggered{false};
-	std::unique_ptr<struct event, std::function<void(struct event*)>> recheck_event{nullptr, event_free};
+	event_ptr recheck_event{nullptr, event_free};
 
 	/* to workaround lack of fuse_buf_free(), keep a global buffer */
 	/* https://github.com/libfuse/libfuse/issues/1373 */
@@ -112,6 +117,7 @@ enum class steve_token_availability {
 	available,
 	no_tokens,
 	load_exceeded,
+	per_process_limit_exceeded,
 };
 
 static void steve_get_load(steve_state *state)
@@ -170,6 +176,12 @@ static steve_token_availability steve_can_give_token(steve_state *state, uint64_
 	/* if there is a token reserved, we give it immediately (even if load is exceeded now) */
 	if (state->processes[pid].token_reserved)
 		return steve_token_availability::available;
+	/* if the per-process limit is reached, do not provide a token */
+	if (state->per_process_limit > 0 && state->processes[pid].tokens_held >= state->per_process_limit)
+		return steve_token_availability::per_process_limit_exceeded;
+	/* if min_jobs is not satisifed, we always give out a token */
+	if (state->jobs < state->min_jobs + state->tokens)
+		return steve_token_availability::available;
 	/* if we have been throttled due to load, wait for it to time out */
 	if (state->recheck_triggered) {
 		assert(state->tokens > 0);
@@ -179,8 +191,6 @@ static steve_token_availability steve_can_give_token(steve_state *state, uint64_
 	if (state->tokens <= 0)
 		return steve_token_availability::no_tokens;
 	if (state->max_load_avg > 0) {
-		if (state->jobs < state->min_jobs + state->tokens)
-			return steve_token_availability::available;
 		steve_get_load(state);
 		if (state->load_avg > state->max_load_avg) {
 			/* trigger a recheck if we don't have one now */
@@ -244,8 +254,16 @@ static void steve_reserve_token(steve_state *state, uint64_t pid)
 static void steve_wake_waiters(steve_state *state)
 {
 	for (auto it = state->waiters.begin(); it != state->waiters.end();) {
-		if (steve_can_give_token(state, it->pid) != steve_token_availability::available)
-			break;
+		steve_token_availability token_available = steve_can_give_token(state, it->pid);
+		switch (token_available) {
+			case steve_token_availability::available:
+				break;
+			case steve_token_availability::per_process_limit_exceeded:
+				++it;
+				continue;
+			default:
+				return;
+		}
 
 		if (fuse_req_t *read_req = std::get_if<fuse_req_t>(&it->handle)) {
 			/* read request */
@@ -322,7 +340,7 @@ static void steve_init(void *userdata, struct fuse_conn_info *)
 
 	state->tokens = state->jobs;
 
-	std::print(stderr, "steve running on /dev/steve for {} jobs\n", state->jobs);
+	std::print(stderr, "steve running on /dev/{} for {} jobs\n", state->dev_name, state->jobs);
 	if (state->max_load_avg > 0) {
 		std::print(stderr, "  tokens will be served with load average < {:.3}\n", state->max_load_avg);
 		std::print(stderr, "  with a recheck timeout of {} s {} us\n",
@@ -330,6 +348,8 @@ static void steve_init(void *userdata, struct fuse_conn_info *)
 	}
 	if (state->min_jobs > 0)
 		std::print(stderr, "  at least {} jobs will be always available\n", state->min_jobs);
+	if (state->per_process_limit > 0)
+		std::print(stderr, "  per-process limit set to {}\n", state->per_process_limit);
 }
 
 static void steve_destroy(void *userdata)
@@ -378,7 +398,7 @@ static void steve_open(fuse_req_t req, struct fuse_file_info *fi)
 
 	if (state->processes.find(fi->fh) != state->processes.end()) {
 		assert(state->processes[fi->fh].pid_fd != -1);
-		assert(state->processes[fi->fh].event);
+		assert(state->processes[fi->fh].pidfd_event);
 	} else {
 		int pid_fd = syscall(SYS_pidfd_open, context->pid, 0);
 		if (pid_fd == -1) {
@@ -387,8 +407,7 @@ static void steve_open(fuse_req_t req, struct fuse_file_info *fi)
 			return;
 		}
 
-		std::unique_ptr<struct event, std::function<void(struct event*)>>
-			pidfd_event{event_new(state->evb, pid_fd, EV_READ|EV_PERSIST, steve_handle_pidfd, state), event_free};
+		event_ptr pidfd_event{event_new(state->evb, pid_fd, EV_READ|EV_PERSIST, steve_handle_pidfd, state), event_free};
 		if (!pidfd_event) {
 			std::print(stderr, "unable to allocate event for pidfd");
 			close(pid_fd);
@@ -403,7 +422,7 @@ static void steve_open(fuse_req_t req, struct fuse_file_info *fi)
 		}
 
 		state->processes[fi->fh].pid_fd = pid_fd;
-		state->processes[fi->fh].event = std::move(pidfd_event);
+		state->processes[fi->fh].pidfd_event = std::move(pidfd_event);
 	}
 
 	fuse_reply_open(req, fi);
@@ -616,6 +635,10 @@ static void steve_ioctl(
 			);
 			fuse_reply_ioctl(req, 0, &dval, sizeof(dval));
 			break;
+		case STEVE_IOC_GET_PER_PROCESS_LIMIT:
+			val = state->per_process_limit;
+			fuse_reply_ioctl(req, 0, &val, sizeof(val));
+			break;
 		case STEVE_IOC_SET_JOBS:
 			if (val == 0)
 				val = sysconf(_SC_NPROCESSORS_ONLN);
@@ -642,6 +665,12 @@ static void steve_ioctl(
 			fuse_reply_ioctl(req, 0, nullptr, 0);
 			steve_wake_waiters(state);
 			break;
+		case STEVE_IOC_SET_PER_PROCESS_LIMIT:
+			state->per_process_limit = val;
+			std::print(stderr, "PID {} set per-process limit to {}\n", fi->fh, state->per_process_limit);
+			fuse_reply_ioctl(req, 0, nullptr, 0);
+			steve_wake_waiters(state);
+			break;
 		case STEVE_IOC_SET_LOAD_AVG:
 			if (dval < 1) {
 				fuse_reply_err(req, EINVAL);
@@ -695,8 +724,19 @@ static void steve_handle_sigusr1(evutil_socket_t, short, void *userdata) {
 static void steve_handle_cuse(evutil_socket_t, short, void *userdata) {
 	steve_state *state = static_cast<steve_state *>(userdata);
 
-	if (fuse_session_receive_buf(state->session, &state->buf) > 0)
-		fuse_session_process_buf(state->session, &state->buf);
+	int rd = fuse_session_receive_buf(state->session, &state->buf);
+	if (rd <= 0 && rd != -EINTR && rd != -EAGAIN) {
+		state->retval = 1;
+		if (rd == 0)
+			std::print(stderr, "EOF from CUSE, exiting (device already taken?)\n");
+		else
+			std::print(stderr, "Reading from CUSE failed: {}", strerror(-rd));
+		if (event_base_loopbreak(state->evb) == -1) {
+			std::print(stderr, "event_base_loopbreak() failed, forcing hard exit");
+			exit(1);
+		}
+	}
+	fuse_session_process_buf(state->session, &state->buf);
 }
 
 static void steve_handle_recheck(evutil_socket_t, short, void *userdata) {
@@ -705,6 +745,30 @@ static void steve_handle_recheck(evutil_socket_t, short, void *userdata) {
 	steve_wake_waiters(state);
 }
 
+static void steve_handle_exit(evutil_socket_t signum, short, void *userdata) {
+	steve_state *state = static_cast<steve_state *>(userdata);
+
+	std::print("Exiting on SIG{}\n", sigabbrev_np(signum));
+	if (event_base_loopbreak(state->evb) == -1) {
+		std::print(stderr, "event_base_loopbreak() failed, forcing hard exit");
+		exit(1);
+	}
+}
+
+static event_ptr steve_new_signal_handler(steve_state *state, int signum, event_callback_fn handler)
+{
+	event_ptr ret{evsignal_new(state->evb, signum, handler, state), event_free};
+	if (!ret) {
+		std::print(stderr, "failed to initialize SIG{} handler", sigabbrev_np(signum));
+		return nullptr;
+	}
+	if (event_add(ret.get(), nullptr) == -1) {
+		std::print(stderr, "failed to enable SIG{} handler", sigabbrev_np(signum));
+		return nullptr;
+	}
+	return ret;
+}
+
 static bool steve_drop_privileges(const char *user) {
 	errno = 0;
 	struct passwd *pw = getpwnam(user);
@@ -734,6 +798,7 @@ static constexpr char steve_usage[] =
 "options:\n"
 "    --help, -h             print this help message\n"
 "    --version, -V          print version\n"
+"\n"
 "    --jobs=JOBS, -j JOBS   jobs to use (default: nproc)\n"
 "    --load-average=LOAD_AVG, -l LOAD_AVG\n"
 "                           do not serve tokens unless load is below LOAD_AVG\n"
@@ -742,11 +807,20 @@ static constexpr char steve_usage[] =
 "                           (fractional down to usec, default: 0.5)\n"
 "    --min-jobs=MIN_JOBS, -m MIN_JOBS\n"
 "                           min. jobs to serve even if load average is exceeded\n"
+"                           (default: 1)\n"
+"    --per-process-limit=LIMIT, -p LIMIT\n"
+"                           max. jobs to serve to a single process\n"
+"\n"
+"    --dev-name=DEV_NAME    override the device name to use (default: steve)\n"
 "    --user=USER, -u USER   drop superuser privileges and switch to USER\n"
 "                           (and its primary group)\n"
 "    --verbose, -v          enable verbose logging\n"
 "    --debug, -d            enable FUSE debug output\n";
 
+struct steve_long_option {
+	static constexpr int dev_name = 256;
+};
+
 static const struct option steve_long_opts[] = {
 	{ "help", no_argument, 0, 'h' },
 	{ "version", no_argument, 0, 'V' },
@@ -754,13 +828,15 @@ static const struct option steve_long_opts[] = {
 	{ "load-average", required_argument, 0, 'l' },
 	{ "load-recheck-timeout", required_argument, 0, 'r' },
 	{ "min-jobs", required_argument, 0, 'm' },
+	{ "per-process-limit", required_argument, 0, 'p' },
+	{ "dev-name", required_argument, 0, steve_long_option::dev_name },
 	{ "user", required_argument, 0, 'u' },
 	{ "verbose", no_argument, 0, 'v' },
 	{ "debug", no_argument, 0, 'd' },
 	{},
 };
 
-static const char *steve_short_opts = "hVj:l:r:m:u:vd";
+static const char *steve_short_opts = "hVj:l:r:m:p:u:vd";
 
 int main(int argc, char **argv)
 {
@@ -779,6 +855,7 @@ int main(int argc, char **argv)
 				return 0;
 			case 'j':
 			case 'm':
+			case 'p':
 				{
 					long jobs_arg;
 					if (!arg_to_long(optarg, &jobs_arg)) {
@@ -789,6 +866,8 @@ int main(int argc, char **argv)
 						state.jobs = jobs_arg;
 					else if (opt == 'm')
 						state.min_jobs = jobs_arg;
+					else if (opt == 'p')
+						state.per_process_limit = jobs_arg;
 					else
 						assert(0 && "not reached");
 				}
@@ -817,6 +896,9 @@ int main(int argc, char **argv)
 			case 'd':
 				debug = true;
 				break;
+			case steve_long_option::dev_name:
+				state.dev_name = optarg;
+				break;
 			default:
 				std::print(stderr, steve_usage, argv[0]);
 				return 1;
@@ -863,8 +945,9 @@ int main(int argc, char **argv)
 	if (user && !steve_drop_privileges(user))
 		return 1;
 
-	const char *dev_name = "DEVNAME=steve";
-	const char *dev_info_argv[] = { dev_name };
+	std::string dev_name_arg{"DEVNAME="};
+	dev_name_arg += state.dev_name;
+	const char *dev_info_argv[] = { dev_name_arg.c_str() };
 	struct cuse_info ci{};
 	ci.dev_info_argc = 1;
 	ci.dev_info_argv = dev_info_argv;
@@ -884,8 +967,7 @@ int main(int argc, char **argv)
 	}
 	state.session = session.get();
 
-	std::unique_ptr<struct event, std::function<void(struct event*)>>
-		cuse_event{event_new(evb.get(), cuse_fd, EV_READ|EV_PERSIST, steve_handle_cuse, &state), event_free};
+	event_ptr cuse_event{event_new(evb.get(), cuse_fd, EV_READ|EV_PERSIST, steve_handle_cuse, &state), event_free};
 	if (!cuse_event) {
 		std::print(stderr, "failed to initialize CUSE handler");
 		return 1;
@@ -895,16 +977,11 @@ int main(int argc, char **argv)
 		return 1;
 	}
 
-	std::unique_ptr<struct event, std::function<void(struct event*)>>
-		sigusr1_event{evsignal_new(evb.get(), SIGUSR1, steve_handle_sigusr1, &state), event_free};
-	if (!sigusr1_event) {
-		std::print(stderr, "failed to initialize SIGUSR1 handler");
-		return 1;
-	}
-	if (event_add(sigusr1_event.get(), nullptr) == -1) {
-		std::print(stderr, "failed to enable SIGUSR1 handler");
-		return 1;
-	}
+	event_ptr sigusr1_event = steve_new_signal_handler(&state, SIGUSR1, steve_handle_sigusr1);
+	std::vector<event_ptr> term_signal_handlers;
+	for (int signum : {SIGHUP, SIGINT, SIGTERM})
+		term_signal_handlers.push_back(
+				steve_new_signal_handler(&state, signum, steve_handle_exit));
 
 	std::string mountpoint = std::format("/dev/fd/{}", cuse_fd);
 	if (fuse_session_mount(session.get(), mountpoint.c_str()) == -1) {
@@ -914,5 +991,5 @@ int main(int argc, char **argv)
 
 	event_base_dispatch(evb.get());
 	fuse_session_unmount(session.get());
-	return 0;
+	return state.retval;
 }
